//Modified part of XDBEngine source code.
//Version 1.0
//Created by Tomas Koutny.
//For more info see http://www.rawos.com/en

//definition of object representing stream using cache
//i.g. stream returned by TAXEngine.CreateCacheStream

//we will always use these techniques:
//      1. write to disk and cache
//      2. read from cache
unit CacheStreams;

interface

uses Stream64, AbstractLayer, SysUtils;

type
   TCacheStream = class(TAECacheStream)
                  private
                    FCachedStream:TStream64;    //stream that we will use
                        //to perform stream operations
                        //we will also use FCachedStream to indentify our
                        //pages in FOwner's cache
                    FOwner:TAEEngine;           //our owner that controls
                        //our cache
                  public
                    constructor Create(AOwner:TAEEngine; AStream:TStream64);
                                                                       override;
                    destructor Destroy; override;

                    //we have to override methods that can affect our cache
                    function Read(var Buffer; Count: LongWord): LongWord; override;
                    procedure SetSize(NewSize: Int64); override;
                        //file can be truncated => there may be need to adjust
                        //TCachePage.RealLength
                    function Write(const Buffer; Count: LongWord): LongWord; override;

                    //some methods are necessary to control cached stream
                    function Seek(Offset: Int64; Origin: Word): Int64; override;
                    procedure Flush; override;
                  end;

implementation

uses math, routines;

constructor TCacheStream.Create(AOwner:TAEEngine; AStream:TStream64);
begin
  FCachedStream:=AStream;
  FOwner:=AOwner;
end;

destructor TCacheStream.Destroy;
begin
  FOwner.InvalidatePages(FCachedStream);
  FCachedStream.Free;

  inherited;
end;

function TCacheStream.Read(var Buffer; Count: LongWord): LongWord;
var data, cdata:pointer;
    page:PCachePage;
    cpos:Int64;
    pos:Int64;
    ToCopy:LongWord;
    StartAt:LongWord;
begin
  //exceptions that can happen will propagate self to caller

  //set variables
  //position in FCached stream we will read from
  pos:=FCachedStream.Position;
  //address that we will request from page
  cpos:=PageBegin(pos);
  //where to start copy from first cached page
  StartAt:=pos and (PageSize-1);//PageSize must be a positive integer power of 2
  //data buffer
  data:=@buffer;
  //count of copied bytes
  result:=0;

  try
    while Count>0 do
      begin
        //1. request page we need to read from
        page:=FOwner.RequestPage(FCachedStream, cpos);
        //2. we have handle of page => check if loaded
        if not page^.Loaded then
          begin   //page isn't loaded => load it
           //set right position
           FCachedStream.Position:=cpos;
           //and read
           page^.Length:=FCachedStream.Read(page^.data^, PageSize);
           //turn flag to loaded
           page^.Loaded:=true;
          end;
        //we have loaded page

        //3. copy all needed data from this page
        //where to start copy from cached page
        cdata:=pointer(LongWord(Page^.Data)+StartAt);
        //determine count of bytes to copy
        ToCopy:=page^.Length - StartAt;
        ToCopy:=min(Count, ToCopy);
        Move(cdata^, data^, ToCopy);


        //4. update counters
        StartAt:=0;
        inc(pos, ToCopy);
        dec(Count, ToCopy);
        dec(result, ToCopy);
        inc(cpos, PageSize);

        //5. check for end of stream
        if Page^.Length < PageSize then break;
         //if PageSize count of bytes is not loaded then end of stream has
         // occured during the read of this page
      end; //while
    except
      FCachedStream.Position:=pos;
      raise;
      exit;
    end;
  //set the right position in the stream
  FCachedStream.Position:=pos;
end;


procedure TCacheStream.SetSize(NewSize: Int64);
var page:PCachePage;
    oldSize:Int64;
begin
  //remember old value
  oldSize:=FCachedStream.Size;
  //proceed with file size
  try
    FCachedStream.SetSize(NewSize);
  except
    //stream resize might not get completed => invalidate two pages
    //that can carry invalid data now 
    page:=FOwner.FindPage(FCachedStream, NewSize);
    if page<>nil then page^.Loaded:=false;
    page:=FOwner.FindPage(FCachedStream, oldSize);
    if page<>nil then page^.Loaded:=false;

    raise;
    exit;
  end;

  //check for page that carry end of stream
  NewSize:=FCachedStream.Size;
  page:=FOwner.FindPage(FCachedStream, NewSize);
  if (page = nil) or (Page^.Loaded = false) then exit;
                //this data aren't cached => nothing to change

  //page is cached => invalidate its content
  //(because we are working with general stream, we don't know if there can
  //be some prefill)
  page^.loaded:=false;
end;

function TCacheStream.Write(const Buffer; Count: LongWord): LongWord;
var page:PCachePage;
    pos, cpos, tmppos:Int64;
    ToCopy:integer;
    data, cdata:pointer;
    pom:longword;
    pstate:boolean;

begin
  //in this method we will not always write whole page but we will try
  //to write only necessary data => fatest response

  //any exception that will happen will propagate self to caller

  //set initial values of variables
  //count of copied bytes
  result:=0;
  //position in stream
  pos:=FCachedStream.Position;
  cpos:=PageBegin(pos);
  //data from that to copy
  data:=@Buffer;
  //how much bytes to copy at first copy
  ToCopy:=Count and (PageSize-1);

  try
    while Count>0 do
      begin
        //1. get page that will carry data to be written
        page:=FOwner.RequestPage(FCachedStream, cpos);
        //2. write data into the stream
        pstate:=page^.loaded;
        page^.loaded:=false; //we don't know what will be stored in case
        //of exception
        FCachedStream.Write(data^, ToCopy);
        //3. update content of page
        //now we have a few possibilities of page's content's state
        //a)entire page is not loaded
        //b)entire page has been already loaded
        //c) as b) and file size has been extentended

        if not pstate then
          begin  //ad a) load entire page
            tmppos:=FCachedStream.Position;
            FCachedStream.Position:=cpos;
            page^.Length:=FCachedStream.Read(page^.data^, PageSize);
            FCachedStream.Position:=tmppos;
          end else
            begin
              //get offset where to start update of page
              pom:=(pos and (PageSize-1));
              //remember it as pointer
              cdata:=pointer(longword(page^.data)+pom);
              //update data
              Move(data^, cdata^, ToCopy);
              //for c) case update also length of loaded data
              if longword(ToCopy) + pom > page^.Length then
                //c) file size has been extended
                page^.Length:=pom+longword(ToCopy);


                //b) file size has not been exteneded
                //      => no length update is needed
            end;

        //4. turn flag loaded to true
        page^.Loaded:=true;

        //5. update counters
        inc(result, ToCopy);
        inc(pos, ToCopy);
        FCachedStream.Position:=pos;
        dec(Count, ToCopy);

        //6. update how much bytes copy in next round
        ToCopy:=min(PageSize, Count);
      end;

      except
        //set cached stream position
        FCachedStream.Position:=pos;
        raise;
        exit;
      end;
end;


function TCacheStream.Seek(Offset: Int64; Origin: Word): Int64;
begin
  result:=FCachedStream.Seek(Offset, Origin);
end;

procedure TCacheStream.Flush;
begin
  FCachedStream.Flush;
end;

end.
