//Modified part of XDBEngine source code.
//Version 1.5
//Created by Tomas Koutny.
//for more info see http://www.rawos.com/xdbe/index.html

//definition of TMainEngine = class(TAEEngine)
//cache uses LRU
unit Main;

interface

uses
   IStream64Cache, AbstractLayer, Windows, Stream64,
   CachePages, BTree;

type
    TMainEngine = class(TAEEngine)
                  private
                    FPagesLeft:cardinal;        //count of pages to be allocated
                    FAvailPages:cardinal;       //count of pages that can be
                        //taken from end of chain without removing any referer
                        //from BTree(FTree)
                        //doesn't include FBackupPage
                    FNaviReferer:TCPageNode;    //with taken of this we will
                        //navigate through FTree => creating and destroying
                        //can slow us down too much
                    FBackupPage:TCachePage;     //RequestPage always must
                                //return non-nil value => one extra page
                                //that must always invalidate after its use
                    procedure TryAllocPages;     //try to alloc memory
                              //for pages until request size of cache is reached
                              //and sets FCacheSize
                              //presumes FCTail to be <>nil (speed option)
                    function AllocChain:PCPageChain;
                              //creates new item into the chain that holds
                              //reference about all cache nodes & pages
                              //doesn't make complete initialization of result
                              //if chain cann't be created then raises exception   
                  protected
                    FCHead, FCTail:PCPageChain; //point to head and tail of
                        //cache pages chain - this will be used to quickly
                        //find last recently used page
                    FTree:TBTree;  //tree to search through cache
                  public
                    constructor Create(ACacheSize:Cardinal);
                         //Allocates at least one page for cache
                         //if ACaheSize <> 0

                    destructor Destroy; override; //frees memory

                    function CreateCacheStream(AStream:TStream64):TStream64; override;
                        //Creates stream that will use cache of TACache64Engine
                        //AStream will control access to medium storing data
                        //and caller must use only returned value
                        //for any further operations
                        //TACache64Engine will automatically free AStream
                        //at call returned.Free
                        //Can raise an exception when error happens
                    function RequestPage(AStream:TStream64; APosition:Int64):
                                                           PCachePage; override;
                     //this function is to be called by TAECacheStream
                     //returns page that holds positions from APosition*PageSize
                     //to APosition*(PageSize+1)-1
                     //nil can not be returned
                    function FindPage(AStream:TStream64; APosition:Int64):
                                                           PCachePage; override;
                      //same meaning as RequestPage but can return nil if
                      //page is not cached => useful for call as TCachedStream.
                      //SetSize that doesn't need to load data into the cache
                    procedure InvalidatePages(AStream:TStream64); override;
                             //invalidates contents of all pages belonging
                             //to the stream AStream
                  end;

implementation

uses SysUtils, CacheStreams, routines, math;


constructor TMainEngine.Create(ACacheSize:Cardinal);
begin
  //initialize cache
  FCHead:=nil;
  FCTail:=nil;
  FTree:=nil;
  FCacheSize:=0;
  FPagesLeft:=0;
  FNaviReferer:=TCPageNode.Create;
  //try to allocate cache memory
  if ACacheSize > 0 then
    begin
      //we need to create at least 2 pages at this time
      FCacheSize:=PageSize shl 1;
      FCacheSize:=max(ACacheSize, FCacheSize);

      //try to alloc two pages(speed up in request page),
      //BackUp page and create BTree
      FBackupPage.Data:=pointer(GlobalAlloc(GMEM_FIXED or GMEM_NODISCARD,
                                PageSize));
      if FBackupPage.Data = nil then Abort;
      FBackupPage.Loaded:=false;

      //alloc two pages
      FCHead:=AllocChain;
      FCTail:=AllocChain;
      FCHead^.next:=FCTail;
      FCTail^.prev:=FCHead;
      FAvailPages:=2;

      //calculate how much pages we need to create by TryAllocPages
      FPagesLeft:=(FCacheSize shr PageSHR) - 2;
      FCacheSize:=PageSize shl 1;
      TryAllocPages;

      FTree:=TBTree.Create(BTreeN);
    end;
end;

destructor TMainEngine.Destroy;
var cache:PCPageChain;
begin
  //free cache
  if FCacheSize > 0 then
    begin
      GlobalFree(cardinal(FBackupPage.Data));
      FTree.Free; //tree must be destroyed before its referers are destroyed

      try            //if any error => invalid chain => leave it as it is
        while FCHead <> nil do
          begin
            cache:=FCHead^.Next;
            FCHead^.node.Free;
            dispose(FCHead);
            FCHead:=cache;
          end;
      except
      end;

      FNaviReferer.FPage:=nil; //referer can not own its page
                        //after some optimalizations - there is some page left
                        //but this page has been already freed up
      FNaviReferer.Free;
    end;
end;


function TMainEngine.CreateCacheStream(AStream:TStream64):TStream64;
begin
  if FCacheSize = 0 then result:=AStream //we haven't own cache this time
    else result:=TCacheStream.Create(Self, AStream);
end;

function TMainEngine.AllocChain:PCPageChain;
begin
  result:=nil;
  try
    new(result);
    with result^ do
      begin
        next:=nil;
        prev:=nil;

        Node:=TCPageNode.Create;
        //we have node allocated => prepare its page
        new(Node.FPage);
        Node.FPage^.Data:=pointer(GlobalAlloc(GMEM_FIXED or GMEM_NODISCARD,
                                  PageSize));
        Node.FPage^.Loaded:=false;
        //FPage's memory will be discarded by Node.Destroy
      end;
    except //chain alloc except
      if result<>nil then
        try    //ensure that no exception will arise
          result^.Node.Free;
        finally
          dispose(result);
        end;
      raise;
  end;  //chain alloc except
end;

procedure TMainEngine.TryAllocPages;
var Chain:PCPageChain;
begin
  chain:=nil;   //suppress message of compiler that variable chain
                //might not have been initialized

  while FPagesLeft > 0 do
    begin
      try
        chain:=AllocChain;
        if chain = nil then Abort;
      except
        exit;
      end;

      //update counters
      dec(FPagesLeft);
      inc(FCacheSize, PageSize);
      inc(FAvailPages);

      //update relations
      chain^.prev:=FCTail;
      FCTail^.Next:=chain;
      FCTail:=chain;
    end;
end;

procedure TMainEngine.InvalidatePages(AStream:TStream64);
var tmp:PCPageChain;
begin
  tmp:=FCHead;
  while tmp<>nil do
    begin
      with tmp^.node do
        if Stream = AStream then FPage^.Loaded:=false;
      tmp:=tmp^.next;
    end;
end;

function TMainEngine.FindPage(AStream:TStream64; APosition:Int64):PCachePage;
var tmp:TCPageNode;
begin
  FNaviReferer.Stream:=AStream;
  FNaviReferer.Position:=PageBegin(APosition);
  tmp:=FTree.Find(FNaviReferer, false) as TCPageNode;
  result:=nil;
  if tmp<>nil then result:=tmp.FPage;
end;

function TMainEngine.RequestPage(AStream:TStream64; APosition:Int64):PCachePage;
var tmpref, oldnavi:TCPageNode;
begin
  //we will try to process page request normally, but if any exception arise
  //then we give to caller our backup page

  //because we have allocated at least two pages, then we count with
  //fact that FCHead<>FCTail in following code

  //try to locate our page
  FNaviReferer.Stream:=AStream;
  FNaviReferer.Position:=PageBegin(APosition);
  //check FCHead to skip tree search - on this step demand some next calls
  with FCHead^.node do
    if Compare(FNaviReferer) = 0 then
      begin //FCHead holds demanded page
        result:=FPage;
        exit;
      end;

  oldnavi:=FNaviReferer;
  try
    //locate demanded page
    tmpref:=FTree.Find(FNaviReferer, true) as TCPageNode;

    //examine if FNaviReferer has been added or not
    if tmpref = FNaviReferer then
      begin //yes, FNaviReferer has been added
        //if we have all pages used then we have to remove one reference
        //from BTree

        //try to alloc memory if needed
        if FAvailPages<1 then TryAllocPages;

        FNaviReferer:=FCTail^.node; //get new navigation referer
        if FAvailPages < 1 then FTree.Delete(FNaviReferer)
             //last recently used page must removed from tree
          else dec(FAvailPages); //we will take one free page

        tmpref.FChain:=FCTail; //free page is always taken from chain tail
        tmpref.FPage:=FCTail^.node.FPage;  //get free memory
        tmpref.FPage^.Loaded:=false; //page has been freshly added & not loaded
        //adjust FCTail to penult chain
        FCTail:=FCTail^.prev;
        FCTail^.next:=nil;
      end;
      //else referer has not been added => only to update chain's elements order

      //move node to chain's begin
      with tmpref.FChain^ do
        begin
          prev^.next:=next;       //by test of FCHead we know that prev<>nil
          if next<>nil then next^.prev:=prev;
          //relations inside of chain fixed => move FChain to begin
          prev:=nil;
          next:=FCHead;
          FCHead^.prev:=tmpref.FChain;
          FCHead:=FCHead^.prev;
          FCHead^.node:=tmpref;

          //give to user demanded page
          result:=node.FPage;
        end;

  except
    FBackupPage.Loaded:=false; //only for crisis situations, not to hold data
    result:=@FBackupPage;

    //tree could be in the inconsistent state => rebuild tree
    FTree.Free;
    FTree:=TBTree.Create(BTreeN);
    FAvailPages:=FCacheSize shr PageSHR;

    //update referer
    FNaviReferer:=oldnavi;
  end;
end;

end.
